"""
This file defines a dummy class html_tag_list, used to provide auto-completion of TreeSeed

Custom tags can be added to the list, and running the script.
# The script will then self modify to update the static class declaration
# following the comment line: # Auto generated class definition
# The static definition is needed in order for autocomplete to pick it up.
"""

tags = 'body title div p h1 h2 h3 h4 h5 h6 u b i s a em strong span font del_ ins ul ol li dd dt dl ' \
       'article section nav aside header footer audio video object_ embed param fieldset legend button textarea ' \
       'label select option table thead tbody tr th td caption blockquote cite q abbr acronym address meta link ' \
       'br hr input_ img code samp pre var kbd dfn' \
       + 'plot datatable graphics html head style script'

def definition_generator(tags):
    classdef = ''

    # This code adds the static class definition
    classdef += 'class HTMLTagList:\n'
    classdef += ' '*4 + '''"""Dummy class used to provide autocomplete for TreeSeed"""\n'''
    for tag in tags.split():
        classdef += ' '*4 + tag + ' = None\n'
    return classdef


def rebuild_content(content):  # pragma: no cover
    # This code reloads the defining code from the file.
    classdef = ''
    with open(__file__) as f:
        for line in f:
            if line.startswith('# Auto generated class definition'):
                break
            else:
                classdef += line
    classdef += '# Auto generated class definition\n'
    classdef += content
    with open(__file__, 'w') as f:
        f.write(definition_generator())


if __name__ == '__main__':  # pragma: no cover
    rebuild_content(definition_generator(tags))


# Auto generated class definition
class HTMLTagList:
    """Dummy class used to provide autocomplete for TreeSeed"""
    body = None
    title = None
    div = None
    p = None
    h1 = None
    h2 = None
    h3 = None
    h4 = None
    h5 = None
    h6 = None
    u = None
    b = None
    i = None
    s = None
    a = None
    em = None
    strong = None
    span = None
    font = None
    del_ = None
    ins = None
    ul = None
    ol = None
    li = None
    dd = None
    dt = None
    dl = None
    article = None
    section = None
    nav = None
    aside = None
    header = None
    footer = None
    audio = None
    video = None
    object_ = None
    embed = None
    param = None
    fieldset = None
    legend = None
    button = None
    textarea = None
    label = None
    select = None
    option = None
    table = None
    thead = None
    tbody = None
    tr = None
    th = None
    td = None
    caption = None
    blockquote = None
    cite = None
    q = None
    abbr = None
    acronym = None
    address = None
    meta = None
    link = None
    br = None
    hr = None
    input_ = None
    img = None
    code = None
    samp = None
    pre = None
    var = None
    kbd = None
    dfnplot = None
    datatable = None
    graphics = None
    html = None
    head = None
    style = None
    script = None



from pyno.tree_model import html as H, TreeSub, TreeNode

# Create a special element for CDATA insertion
class CDATA(TreeSub):
    def construct(self, content, *args, **kwargs):
        return f"<![CDATA[{content}]]>"


# Create a special element that works for including different asset types. eg. javascript and stylesheets
class Include(TreeSub):
    def construct(self, file_path, *args, **kwargs):
        if file_path.endswith('.js'):
            return H.script(type="text/javascript", src=file_path)
        elif file_path.endswith('.css'):
            return H.link(rel="stylesheet", type="text/css", href=file_path)
        else:
            raise ValueError('Included file ending is not supported in H.Include' + file_path)


class Import(TreeSub):
    def construct(self, file_path, *args, **kwargs):
        if file_path.endswith('.js'):
            return H.script(type="text/javascript", src=file_path)
        elif file_path.endswith('.css'):
            return H.link(rel="stylesheet", type="text/css", href=file_path)
        else:
            raise ValueError('Included file ending is not supported in H.Include' + file_path)
        # todo consider adding more as it becomes relevant to usage



# todo update docs, defaults are acquired at definition time. Show trick for doing late bound attributes

import collections
from functools import partial
import inspect

def get_default_args(func):
    return {
        k: v.default
        for k, v in inspect.signature(func).parameters.items()
        if v.default is not inspect.Parameter.empty}


class ClassyMaker:
    """This is a wrapper class that provides the syntax H.div['class'] and H.div.class
    for defining a class while calling the element"""
    def __init__(self, node):
        self.node = node
        self.Class = []

    def __getattr__(self, item):
        self.Class.append(item)
        return self

    def __getitem__(self, item):
        self.Class.append(item)
        return self

    def __call__(self, *args, **kwargs):
        return self.node(*args, **{k: v for k, v in kwargs.items() if k.lower() != 'class'},
                         Class=' '.join(x.replace('_', '-') for x in [*self.Class, kwargs.get('class'), kwargs.get('Class', '')] if x is not None))


class TreeNode:
    """tree_node is an object used to construct object-trees for generation of structed text like html/xhtml/svg code"""

    # defaults = {}  # Contains default values used for tag properties

    def __new__(typ, _tagname, *args, namespace='', Root=None, **kwargs):
        obj = object.__new__(typ)
        obj.namespace = namespace
        obj.name = _tagname
        obj.args = list(args)
        # The kwargs are initiated using defaults for the tag if they exist
        if Root is None:
            Root = HTML
        obj.kwargs = dict(Root.defaults[obj.name].copy(), **kwargs) if obj.name in Root.defaults else kwargs
        # todo perhaps given arguments should overwrite the defaults... otherwise they are more a sort of superfaults
        return obj

    def __getattr__(self, item):
        if item in object.__getattribute__(self, 'kwargs').keys():
            return self.kwargs[item]
        else:
            return object.__getattribute__(self, item)

    def __setattr__(self, key, value):
        if key not in ('name', 'args', 'kwargs', 'value', 'namespace'):
            self.kwargs[key] = value
        else:
            object.__setattr__(self, key, value)

    def __str__(self):
        return str(self.construct(*self.args, **self.kwargs))

    def construct(self, *args, **kwargs):
        """ This function constructs the output format. Here it's building a string representation of html"""
        # Properties with _ preceding are not passed on to html.
        property_args = {k: v for k, v in self.kwargs.items() if not k.startswith('_') and not (k == 'Class' and len(v.strip())==0)}

        # Generate attribute definitions:
        def value_exp(value):
            if value is None:
                return ''
            else:
                return f'="{value}"'
        properties = (' '+' '.join([f'{name.replace("_", "-")}{value_exp(value)}' for name, value in property_args.items()])) \
            if len(property_args) > 0 else ''

        # This raplaces iterators with lists so they aren't exhausted if the structure is iterated more than once.
        for n, elm in enumerate(self.args):
            if isinstance(elm, collections.Iterator):
                self.args[n] = list(self.args[n])

        # Generate content string
        content_string = ''.join([''.join(str(x) for x in line) if hasattr(line, '__iter__')
                                  else str(line) for line in self.args])

        # Return content with enclosing tag
        if self.kwargs.get('_void_element', False):
            return f'<{self.namespace}{self.name}{properties} />'
        else:
            return f'<{self.namespace}{self.name}{properties}>{content_string}</{self.namespace}{self.name}>'


    def __call__(self, environ, start_response, **kwargs): # pragma: no cover
        """ When called a TreeNode object will act like a WSGI application.
        This way they can be passed directly as output in for instance Flask apps, and served using the wsgi
        reference implementation easily """
        start_response('200 OK', [('Content-type', 'text/HTML charset=utf-8')])
        return [str(self)]


class NodeDispatcher(type):
    """ Diverts attributes to subclasses when they exist,
    and includes default arguments of construct in the initializer
    """
    def __getattr__(self, attr):
        for sub_node in HTML.__subclasses__():
            if sub_node.__name__ == attr:
                # todo adding default arguments cause 3x-4x slowdown
                return ClassyMaker(partial(sub_node, **get_default_args(sub_node.construct)))
                # return sub_node
        else:
            return ClassyMaker(partial(TreeNode, attr))


class HTML(TreeNode, metaclass=NodeDispatcher):  # type: HTMLTagList
    """ HTML is a class used to:
    * create TreeNodes though attribute dispatching     eg. H.div())
    * register user-defined tags through subclassing    eg. class myclass(H)
    * dispatch to user-defined classes though attribute dispatching eg. H.myclass()
    * Track default arguments to nodes eg. H.defaults['myclass'] = {'a': 3, 'b': 3}
    """

    @staticmethod
    def construct(func, **kwargs):
        """ This static method is a decorator that allows constructing subclass by decorating construct functions.
        This simplifies the main use case of construct-classes which is simply to wrap some structure without any class state or additional methods management"""
        globals()[func.__name__] = type(func.__name__, (HTML,), {"construct": lambda self, *args, **kwargs: func(*args, **kwargs)})
        return func

    defaults = {}

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, cls.__name__, *args, **kwargs)
    pass


class NodeDispatcherNS(type):
    """ Diverts attributes to subclasses when they exist,
    and includes default arguments of construct in the initializer
    """

    def __getattr__(self, attr):
        for sub_node in Namespace.__subclasses__():
            if sub_node.__name__ == attr:
                # todo adding default arguments cause 3x-4x slowdown
                return ClassyMaker(partial(sub_node, namespace=self._namespace, Root=self, **get_default_args(sub_node.construct)))
        else:
            return ClassyMaker(partial(TreeNode, attr, namespace=self._namespace, Root=self))


# todo is it still appropriate to use a metaclass pathway to attribute node generation considering the class suddenly has a namespace argument?

class Namespace(TreeNode, metaclass=NodeDispatcherNS):  # type: HTMLTagList
    """ HTML is a class used to:
    * create TreeNodes though attribute dispatching     eg. H.div())
    * register user-defined tags through subclassing    eg. class myclass(H)
    * dispatch to user-defined classes though attribute dispatching eg. H.myclass()
    * Track default arguments to nodes eg. H.defaults['myclass'] = {'a': 3, 'b': 3}
    """
    _namespace = ''
    defaults = {}

    @staticmethod
    def construct(func, **kwargs):
        """ This static method is a decorator that allows constructing subclass by decorating construct functions.
        This simplifies the main use case of construct-classes which is simply to wrap some structure without any class state or additional methods management"""
        globals()[func.__name__] = type(func.__name__, (HTML,), {"construct": lambda self, *args, **kwargs: func(*args, **kwargs)})
        return func


    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, cls.__name__, *args, **kwargs)
    pass



"""html is an instantiated object of the TreeSeed class, 
used to provide easy generation of TreeNodes through attribute access 
Type annotation against HTMLTagList is added purely to be able to have autocompletion in editors that support this"""
from pyno.html_tags_autocomplete import HTMLTagList
# Backwards compatibility
html = HTML  # type: HTMLTagList
TreeSub = HTML
